//
// Creates Hamiltonian circuit programs.
//
#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <stdlib.h>

extern "C"
{
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_rand.h>
#include <gb_flip.h>
#include <gb_plane.h>
}

Graph *makeLadder(long steps);
void printHamilton(Graph *g);
void printHamilton_new(Graph *g);
void printHamilton_new_dlv(Graph *g);
void printHamilton_cnf(Graph *g);
void printHamilton_cnf2(Graph *g);
void printHamilton2(Graph *g);
void printFacts(Graph *g);

int
main (int argc, char *argv[])
{
  Graph *g;
  
  bool error = false;
  bool ladder = false;
  bool newtrans = false;
  bool cnf = false;
  bool cnf2 = false;
  bool facts = false;
  bool dlv = false;
  long seed = 0;
  char *endptr = "";
  unsigned long nodes = 2;
  double remove = 0;
  for (int c = 1; c < argc && !error; c++)
    {
      if (argv[c][0] == '-')
	{
	  if (strcmp (&argv[c][1], "nodes") == 0)
	    {
	      c++;
	      if (c < argc)
		nodes = strtoul (argv[c], &endptr, 0);
	      else
		error = true;
	    }
	  else if (strcmp (&argv[c][1], "seed") == 0)
	    {
	      c++;
	      if (c < argc)
		seed = strtol (argv[c], &endptr, 0);
	      else
		error = true;
	    }
	  else if (strcmp (&argv[c][1], "remove") == 0)
	    {
	      c++;
	      if (c < argc)
		remove = strtod (argv[c], &endptr);
	      else
		error = true;
	    }
	  else if (strcmp (&argv[c][1], "ladder") == 0)
	    ladder = true;
	  else if (strcmp (&argv[c][1], "new") == 0)
	    newtrans = true;
	  else if (strcmp (&argv[c][1], "cnf") == 0)
	    cnf = true;
	  else if (strcmp (&argv[c][1], "cnf2") == 0)
	    cnf2 = true;
	  else if (strcmp (&argv[c][1], "facts") == 0)
	    facts = true;
	  else if (strcmp (&argv[c][1], "dlv") == 0)
	    dlv = true;
	  else
	    error = true;
	  if (*endptr != '\0')
	    error = true;
	}
      else
	error = true;
    }
  if (nodes <= 1 || remove < 0 || remove > 1)
    error = true;
  if (error)
    {
      cerr << "Usage:  hamilton [-ladder] <-nodes number of nodes>"
	   << " [-seed seed]"
	   << "                 [-new] [-cnf] [-remove fraction]" << endl;
      cerr << "        -new gives a new-rules translation" << endl;
      cerr << "        -cnf prints a sat-problem" << endl;
      cerr << "        -facts prints the graph as facts" << endl;
      return 1;
    }
  
  gb_init_rand(1);
  for (int i = 0; i < 10; i++)
    gb_next_rand();
  if (ladder)
    g = makeLadder(nodes/2);
  else
    g = plane(nodes,0,0,0,(unsigned long)(remove*0xffff),seed);
  if (panic_code)
    cerr << "Graph creation error: " << panic_code << endl;
  if (g == NULL)
    {
      cerr << "Couldn't create graph." << endl;
      return 1;
    }
  if (facts)
    printFacts (g);
  else if (cnf)
    printHamilton_cnf (g);
  else if (cnf2)
    printHamilton_cnf2 (g);
  else if (newtrans)
    printHamilton_new (g);
  else if (dlv)
    printHamilton_new_dlv (g);
  else
    printHamilton (g);
  gb_recycle (g);
}

Graph *makeLadder(long steps)
{
  Graph *g = gb_new_graph (steps*2);
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return 0;
    }
  long i;
  char buf[11];
  for (i = 0; i < steps*2; i++)
    {
      sprintf(buf, "%ld", i);
      g->vertices[i].name = gb_save_string(buf);
    }
  for (i = 0; i < steps*2-2; i += 2)
    {
      gb_new_edge(g->vertices+i, g->vertices+i+1, 1);
      gb_new_edge(g->vertices+i, g->vertices+i+2, 1);
      gb_new_edge(g->vertices+i+1, g->vertices+i+3, 1);
    }
  gb_new_edge(g->vertices+i, g->vertices+i+1, 1);
  return g;
}

void
printHamilton(Graph *g)
{ 
  Graph *h;
  h = gb_new_graph (g->n);  // Reverse graph.
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  long i;
  Arc *a;
  Vertex *base;
  //  cout << "w0 :- not w0." << endl;
  //  cout << "s :- w0." << endl;
  cout << "v0." << endl;
  for (i = 0; i < g->n; i++)
    {
      g->vertices[i].u.I = i;
      h->vertices[i].u.I = i;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      if (i)
	{
	  //	  cout << 'v' << base->name << " :- " << "not v" << base->name << '.' 
	  //	       << endl;
	  //	  cout << "s :- v" << base->name << '.' << endl;
	}
      a = base->arcs;
      while (a)
	{
	  gb_new_arc(&(h->vertices[a->tip->u.I]), &(h->vertices[i]), 1);
	  if (a->tip->name[0] == '0')
	    cout << 'w' << a->tip->name << " :- v" << base->name
		 << ",v" << base->name << 'v' << a->tip->name << '.' << endl;
	  else
	    cout << 'v' << a->tip->name << " :- v" << base->name
		 << ",v" << base->name << 'v' << a->tip->name << '.' << endl;
	  a = a->next;
	}
    }
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      a = base->arcs;
      while (a)
	{
	  cout << 'v' << base->name << 'v' << a->tip->name 
	       << " :- not v" << a->tip->name << 'v' << base->name;
	  Arc *b = h->vertices[a->tip->u.I].arcs;
	  while (b)
	    {
	      if (b->tip->u.I != i)
		cout << ",not v" << g->vertices[b->tip->u.I].name 
		     << 'v' << a->tip->name;
	      b = b->next;
	    }
	  b = base->arcs;
	  while (b)
	    {
	      if (a->tip->u.I != b->tip->u.I)
		cout << ",not v" << base->name << 'v' << b->tip->name;
	      b = b->next;
	    }
	  cout << '.' << endl;
	  a = a->next;
	}
    }
  cout << "compute { w0";
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      cout << ", " << 'v' << base->name;
    }
  cout << " }" << endl;
  gb_recycle (h);
}

void
printArc (Vertex *base, Vertex *tip)
{
  char *b = base->name;
  char *t = tip->name;
  if (strcmp (b, t) > 0)
    {
      char *tmp = b;
      b = t;
      t = tmp;
    }
  cout << 'v' << b << 'v' << t;
}

int
printArcs(Vertex *base, bool not)
{
  bool comma = false;
  int n = 0;
  for (Arc *a = base->arcs; a; a = a->next)
    {
      if (comma == false)
	comma = true;
      else
	cout << ',';
      if (not)
	cout << " not";
      cout << ' ';
      printArc (base, a->tip);
      n++;
    }
  return n;
}

void
printHamilton_new(Graph *g)
{ 
  long i;
  Arc *a;
  Vertex *base;
  cout << "v0." << endl;
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      for (a = base->arcs; a; a = a->next)
	{
	  if (a->tip->name[0] == '0')
	    continue;
	  /*	    cout << 'w';
		    else*/
	    cout << 'v';
	  cout << a->tip->name << " :- v" << base->name << ", ";
	  printArc (base, a->tip);
	  cout << '.' << endl;
	}
    }
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      if (base->arcs == 0)
	continue;
      cout << '{';
      int n = printArcs (base, false);
      cout << " }." << endl;
      cout << "false :- 3 {";
      printArcs (base, false);
      cout << " }." << endl;
      if (n == 1)
	continue;
      cout << "false :- " << n-1 << " {";
      printArcs (base, true);
      cout << " }." << endl;
    }
  cout << "compute { not false";
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      cout << ", " << 'v' << base->name;
    }
  cout << " }" << endl;
}

void
printHamilton_new_dlv(Graph *g)
{ 
  long i;
  Arc *a;
  Vertex *base;
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      for (a = base->arcs; a; a = a->next)
	{
	  if (a->tip->name[0] == '0')
	    continue;
	  /*	    cout << 'w';
		    else*/
	    cout << 'v';
	  cout << a->tip->name << " :- ";
	  if (base->name[0] != '0')
	    cout << 'v' << base->name << ", ";
	  printArc (base, a->tip);
	  cout << '.' << endl;
	}
    }
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      if (base->arcs == 0)
	continue;
      int n = 0;
      for (Arc *a = base->arcs; a; a = a->next)
	{
	  printArc (base, a->tip);
	  cout << " :- not n";
	  printArc (base, a->tip);
	  cout << '.' << endl;
	  cout << 'n';
	  printArc (base, a->tip);
	  cout << " :- not ";
	  printArc (base, a->tip);
	  cout << '.' << endl;
	  n++;
	}
      if (n > 2)
	for (int i = 0; i < n; i++)
	  for (int j = i+1; j < n; j++)
	    for (int k = j+1; k < n; k++)
	      {
		int l = 0;
		cout << " :- ";
		bool comma = false;
		for (Arc *a = base->arcs; a; a = a->next, l++)
		  {
		    if (l != i && l != j && l != k)
		      continue;
		    if (comma == false)
		      comma = true;
		    else
		      cout << ',';
		    printArc (base, a->tip);
		  }
		cout << '.' << endl;
	      }
      if (n == 1)
	continue;
      for (int i = 0; i < n; i++)
	{
	  int j = 0;
	  cout << " :- ";
	  bool comma = false;
	  for (Arc *a = base->arcs; a; a = a->next, j++)
	    {
	      if (i == j)
		continue;
	      if (comma == false)
		comma = true;
	      else
		cout << ',';
	      cout << "not ";
	      printArc (base, a->tip);
	    }
	  cout << '.' << endl;
	}
    }
  for (i = 1; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      cout << " :- " << "not v" << base->name << '.' << endl;
    }
}

void
printHamilton_cnf(Graph *g)
{
  Graph *h = complement (g, 0, 0, 0); // Complement graph
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  for (int i = 0; i < h->n; i++)
    h->vertices[i].u.I = i;
  cout << "c Find Hamiltonian cycle in random planar graph, nodes="
       << g->n << endl;
  int atoms = g->n*g->n;
  int arcs = 0;
  for (int i = 0; i < h->n; i++)
    {
      Vertex *base = &(h->vertices[i]);
      for (Arc *a = base->arcs; a; a = a->next)
	arcs++;
    }
  if (arcs % 2)
    cerr << "error, odd number of arcs" << endl;
  int clauses = g->n*g->n*(g->n-1)+2*g->n+g->n*arcs;
  cout << "p cnf " << atoms << ' ' << clauses << endl;
  // Make permutations
  for (int i = 0; i < g->n; i++)
    {
      // x_i1|...|x_in      n
      cout << i*g->n+1;
      for (int j = 1; j < g->n; j++)
	cout << ' ' << i*g->n+j+1;
      cout << " 0" << endl;
      clauses--;
      // -x_ij|-x_ik        n*n*(n+1)/2
      for (int j = 0; j < g->n; j++)
	for (int k = j+1; k < g->n; k++)
	  {
	    cout << -(i*g->n+j+1) << ' ' << -(i*g->n+k+1) << " 0" << endl;
	    clauses--;
	  }
      // x_1i|...|x_ni      n
      cout << i+1;
      for (int j = 1; j < g->n; j++)
	cout << ' ' << j*g->n+i+1;
      cout << " 0" << endl;
      clauses--;
      // -x_ji|-x_ki        n*n*(n+1)/2
      for (int j = 0; j < g->n; j++)
	for (int k = j+1; k < g->n; k++)
	  {
	    cout << -(j*g->n+i+1) << ' ' << -(k*g->n+i+1) << " 0" << endl;
	    clauses--;
	  }
    }
  // Deny not adjacent 
  for (int i = 0; i < h->n; i++)
    {
      Vertex *base = &(h->vertices[i]);
      for (Arc *a = base->arcs; a; a = a->next)
	{
	  // -x_ik|-x_jk+1  n*(n*(n-1)-arcs)
	  int j = a->tip->u.I;
	  for (int k = 0; k < h->n; k++)
	    {
	      cout << -(i*h->n+k+1) << ' ' 
		   << -(j*h->n+((k+1)%h->n)+1) << " 0" << endl;
	      clauses--;
	    }
	}
    }
  if (clauses)
    cerr << "error, atoms=" << atoms << " clauses=" << clauses << endl;
  gb_recycle (h);
}

bool
print_atleast_cnf (int *neighbour, int n, int k, int t, bool pos, int v)
{
  bool print = false;
  for (int i = t; i <= n-k; i++)
    {
      if (print)
	{
	  print = false;
	  for (int j = 0; j < n; j++)
	    {
	      if (neighbour[j] >= 0)
		continue;
	      int a = v;
	      int b = -neighbour[j]-1;
	      if (a > b)
		{
		  a = b;
		  b = v;
		}
	      if (!pos)
		cout << "not ";
	      cout << "edge(" << a << ',' << b << "), ";
	    }
	}
      else if (t)
	cout << ", ";
      if (!pos)
	cout << "not ";
      int a = v;
      int b = neighbour[i];
      if (a > b)
	{
	  a = b;
	  b = v;
	}
      cout << "edge(" << a << ',' << b << ")";
      neighbour[i] = -neighbour[i]-1;
      if (k-1 == 0)
	{
	  cout << '.' << endl;
	  print = true;
	}
      else
	print = print_atleast_cnf (neighbour, n, k-1, i+1, pos, v);
      neighbour[i] = -neighbour[i]-1;
    }
  return print;
}

void
printHamilton_cnf2(Graph *g)
{
  int *neighbour = new int[g->n];
  for (int i = 0; i < g->n; i++)
    g->vertices[i].u.I = i;
  // Exactly two edges
  for (int i = 0; i < g->n; i++)
    {
      Vertex *base = &(g->vertices[i]);
      int k = 0;
      for (Arc *a = base->arcs; a; a = a->next)
	{
	  int j = a->tip->u.I;
	  neighbour[k++] = j;
	}
      print_atleast_cnf (neighbour, k, k-2+1, 0, true, i);
      print_atleast_cnf (neighbour, k, k-(k-2)+1, 0, false, i);
    }
  // Reachable
  for (int i = 0; i < g->n; i++)
    {
      for (int l = 1; l < g->n; l++)
	cout << "reach(" << i << ',' << l << "), ";
      cout << "reach(" << i << ',' << g->n << ")." << endl;
      Vertex *base = &(g->vertices[i]);
      int k = 0;
      for (Arc *a = base->arcs; a; a = a->next)
	{
	  int j = a->tip->u.I;
	  neighbour[k++] = j;
	}
      for (int l = 0; l < g->n; l++)
	for (int j = 0; j < k; j++)
	  {
	    int a = i;
	    int b = neighbour[j];
	    if (a > b)
	      {
		a = b;
		b = i;
	      }
	    cout << "tmp(" << neighbour[j] << ',' << i << ',' << l
		 << "), " 
		 << "not reach(" << neighbour[j] << ',' << l << "), "
		 << "not edge(" << a << ',' << b << ")." << endl;
	    cout << "not tmp(" << neighbour[j] << ',' << i << ','
		 << l << "), "
		 << "reach(" << neighbour[j] << ',' << l << ")."
		 << endl;
	    cout << "not tmp(" << neighbour[j] << ',' << i << ','
		 << l << "), "
		 << "edge(" << a << ',' << b << ")." << endl;
	  }
      for (int l = 0; l < g->n; l++)
	for (int j = 0; j < k; j++)
	  cout << "reach("<< i << ',' << l+1 << "), " 
	       << "not tmp (" << neighbour[j] << ',' << i << ',' << l
	       << ")." << endl;
      for (int l = 0; l < g->n; l++)
	{
	  for (int j = 0; j < k; j++)
	    cout << "tmp(" << neighbour[j] << ',' << i << ','
		 << l << "), ";
	  cout << "not reach(" << i << ',' << l+1 << ")." << endl;
	}
    }
  cout << "reach(0,0)." << endl;
  for (int i = 1; i < g->n; i++)
    cout << "not reach(" << i << ",0)." << endl;
  delete[] neighbour;
}

void
printHamilton2(Graph *g)
{ 
  Graph *h;
  h = gb_new_graph (g->n);  // Reverse graph.
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  long i;
  Arc *a;
  Vertex *base;
  cout << "w(0) :- not w(0)." << endl;
  cout << "s :- w(0)." << endl;
  cout << "v(0)." << endl;
  for (i = 0; i < g->n; i++)
    {
      g->vertices[i].u.I = i;
      h->vertices[i].u.I = i;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      if (i)
	{
	  cout << "v(" << base->name << ')' << " :- " << "not v(" <<
	    base->name << ")." << endl;
	  cout << "s :- v(" << base->name << ")." << endl;
	}
      a = base->arcs;
      while (a)
	{
	  gb_new_arc(&(h->vertices[a->tip->u.I]), &(h->vertices[i]), 1);
	  if (a->tip->name[0] == '0')
	    cout << "w(" << a->tip->name << ") :- v(" << base->name
		 << "),e(" << base->name << "," << a->tip->name <<
	      ")." << endl;
	  else
	    cout << "v(" << a->tip->name << ") :- v(" << base->name
		 << "),e(" << base->name << ',' << a->tip->name <<
	      ")." << endl;
	  a = a->next;
	}
    }
  if (gb_trouble_code)
    {
      cerr << "Couldn't create graph." << endl;
      return;
    }
  for (i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      a = base->arcs;
      while (a)
	{
	  cout << "e(" << base->name << ',' << a->tip->name 
	       << ") :- not e(" << a->tip->name << ',' << base->name
	       << ')';
	  Arc *b = h->vertices[a->tip->u.I].arcs;
	  while (b)
	    {
	      if (b->tip->u.I != i)
		cout << ",not e(" << g->vertices[b->tip->u.I].name 
		     << ',' << a->tip->name << ')';
	      b = b->next;
	    }
	  b = base->arcs;
	  while (b)
	    {
	      if (a->tip->u.I != b->tip->u.I)
		cout << ",not e(" << base->name << ',' << b->tip->name
		     << ')';
	      b = b->next;
	    }
	  cout << '.' << endl;
	  a = a->next;
	}
    }
  //  cout << "compute { w0";
  //  for (i = 0; i < g->n; i++)
  //    {
  //      base = &(g->vertices[i]);
  //      cout << ", " << 'v' << base->name;
  //    }
  //  cout << " }" << endl;
  gb_recycle (h);
}

void printFacts(Graph *g)
{ 
  Arc *a;
  Vertex *base;
  for (long i = 0; i < g->n; i++)
    {
      g->vertices[i].u.I = i;
      cout << "vertex(" << i << ")." << endl;
    }
  for (long i = 0; i < g->n; i++)
    {
      base = &(g->vertices[i]);
      a = base->arcs;
      for (a = base->arcs; a; a = a->next)
	cout << "arc(" << base->u.I << ',' 
	     << a->tip->u.I << ")." << endl;
    }
}
