#include <stdio.h>

int cnftosm(long atoms, long clauses);
int cnftonewsm(long atoms, long clauses);
int cnftonew2sm(long atoms, long clauses);
int ftosm();
int ftonewsm();
int ftonew2sm();

const size_t l = 4096;
char s[l];

void usage ()
{
  fprintf(stderr, "Usage: sat2sm [-new|-new2]\n");
  fprintf(stderr, "       -new uses a new rules translation.\n");
  fprintf(stderr, "       -new2 uses a complex new rules translation.\n");
}

int main (int argc, char *argv[])
{
  int other = 0;
  int newrules = 0;
  long atoms, clauses;

  if (argc > 2)
    {
      usage ();
      return 1;
    }
  else if (argc == 2)
    {
      if (strcmp(argv[1], "-new") == 0)
	newrules = 1;
      else if (strcmp(argv[1], "-new2") == 0)
	newrules = 2;
      else
	{
	  usage ();
	  return 1;
	}
    }
  fgets(s, l, stdin);
  if (s[0] == 'p' || s[0] == 'c')
    {
      while (s[0] == 'c')
	fgets(s, l, stdin);
      if (sscanf(s, "p cnf %ld %ld\n", &atoms, &clauses) == 2)
	{
	  if (newrules == 1)
	    return cnftonewsm(atoms, clauses);
	  else if (newrules == 2)
	    return cnftonew2sm(atoms, clauses);
	  else
	    return cnftosm(atoms, clauses);
	}
      else
	{
	  fprintf(stderr, "Not in cnf form\n");
	  return 1;
	}
    }
  else if (newrules == 1)
    return ftonewsm();      // Not in cnf form
  else if (newrules == 2)
    return ftonew2sm();      // Not in cnf form
  else
    return ftosm();      // Not in cnf form

}

int cnftosm(long atoms, long clauses)
{
  for (long i = 1; i <= atoms; i++)
    {
      printf("p%ld :- not n%ld.\n", i, i);
      printf("n%ld :- not p%ld.\n", i, i);
    }
  long a;
  for (long i = 1; i <= clauses; i++)
    {
      for (;;)
	{
	  if (scanf("%ld", &a) != 1)
	    {
	      fprintf(stderr, "Error in line %ld\n", i+1);
	      return 1;
	    }
	  if (a == 0)
	    break;
	  if (a > 0)
	    printf("c%ld :- p%ld.\n", i, a);
	  else
	    printf("c%ld :- n%ld.\n", i, -a);
	}
    }
  printf("compute { c1 ");
  for (long i = 2; i <= clauses; i++)
    printf(", c%ld", i);
  printf(" }\n");
}

int cnftonewsm(long atoms, long clauses)
{
  printf ("{ ");
  for (long i = 1; i < atoms; i++)
    printf("a%ld, ", i);
  printf("a%ld }.\n", atoms);
  long a;
  for (long i = 1; i <= clauses; i++)
    {
      bool comma = false;
      printf("false :- ");
      for (;;)
	{
	  if (scanf("%ld", &a) != 1)
	    {
	      fprintf(stderr, "Error in line %ld\n", i+1);
	      return 1;
	    }
	  if (a == 0)
	    break;
	  if (comma)
	    printf(", ");
	  else
	    comma = true;
	  if (a > 0)
	    printf("not a%ld", a);
	  else
	    printf("a%ld", -a);
	}
      printf(".\n");
    }
  printf("compute { not false }\n");
}

int cnftonew2sm(long atoms, long clauses)
{
  int size = 32;
  long *clause = new long[size];
  for (long i = 1; i <= clauses; i++)
    {
      long a = 1;
      bool pos = false;
      bool neg = false;
      for (long j = 0; a; j++)
	{
	  if (scanf("%ld", &a) != 1)
	    {
	      fprintf(stderr, "Error in line %ld\n", i+1);
	      return 1;
	    }
	  if (j >= size)
	    {
	      long *t = new long[2*size];
	      for (int k = 0; k < size; k++)
		t[k] = clause[k];
	      size *= 2;
	      delete[] clause;
	      clause = t;
	    }
	  clause[j] = a;
	  if (a > 0)
	    pos = true;
	  else
	    neg = true;
	}
      if (pos == false && neg == false)
	continue;
      bool comma = false;
      printf("false :- ");
      for (long j = 0; clause[j]; j++)
	{
	  if (comma)
	    printf(", ");
	  else
	    comma = true;
	  if (clause[j] > 0)
	    printf("not a%ld", clause[j]);
	  else
	    printf("a%ld", -clause[j]);
	}
      printf(".\n");
      if (pos)
	{
	  printf("{ ");
	  comma = false;
	  for (long j = 0; clause[j]; j++)
	    {
	      a = clause[j];
	      if (a > 0)
		{
		  if (comma)
		    printf(", ");
		  printf("a%ld", a);
		  comma = true;
		}
	    }
	  printf(" }");
	  if (neg)
	    {
	      printf(" :- ");
	      comma = false;
	      for (long j = 0; clause[j]; j++)
		{
		  a = clause[j];
		  if (a < 0)
		    {
		      if (comma)
			printf(", ");
		      printf("a%ld", -a);
		      comma = true;
		    }
		}
	    }
	  printf(".\n");
	}
    }
  printf("compute { not false }\n");
}

int ftosm()
{
  long maxatom = 0;
  long clauses = 1;
  long a;
  char *t = s;
  long line = 1;
  for (; s[0] != '%';)
    {
      if (sscanf(t, "%ld", &a) != 1)
	{
	  fprintf(stderr, "Error in line %ld\n", line); 
	  return 1;
	}
      if (a > 0)
	{
	  printf("c%ld :- p%ld.\n", clauses, a);
	  if (a > maxatom)
	    maxatom = a;
	}
      else
	{
	  printf("c%ld :- n%ld.\n", clauses, -a);
	  if (-a > maxatom)
	    maxatom = -a;
	}
      while (*t != ' ' && *t != '\0')
	t++;
      if (*t == '\0')
	{
	  clauses++;
	  line++;
	  fgets(s, l, stdin);
	  t = s;
	}
      else
	t++;
    }
  for (long i = 1; i <= maxatom; i++)
    {
      printf("p%ld :- not n%ld.\n", i, i);
      printf("n%ld :- not p%ld.\n", i, i);
    }
  printf("compute { c1 ");
  for (long i = 2; i < clauses; i++)
    printf(", c%ld", i);
  printf(" }\n");
}

int ftonewsm()
{
  long maxatom = 0;
  long a;
  char *t = s;
  long line = 1;
  bool comma = false;
  bool newrule = true;
  for (; s[0] != '%';)
    {
      if (sscanf(t, "%ld", &a) != 1)
	{
	  fprintf(stderr, "Error in line %ld\n", line); 
	  return 1;
	}
      if (newrule)
	{
	  printf("false :- ");
	  newrule = false;
	}
      if (comma)
	printf(", ");
      else
	comma = true;
      if (a > 0)
	{
	  printf("not a%ld", a);
	  if (a > maxatom)
	    maxatom = a;
	}
      else
	{
	  printf("a%ld", -a);
	  if (-a > maxatom)
	    maxatom = -a;
	}
      while (*t != ' ' && *t != '\0')
	t++;
      if (*t == '\0')
	{
	  line++;
	  fgets(s, l, stdin);
	  t = s;
	  printf(".\n");
	  newrule = true;
	  comma = false;
	}
      else
	t++;
    }
  printf("{ ");
  for (long i = 1; i < maxatom; i++)
    printf("a%ld, ", i);
  printf("a%ld }.\n", maxatom);
  printf("compute { not false }\n");
}

int ftonew2sm()
{
  fprintf(stderr, "Old format not implemented!\n");
}
