% Hanoi towers
% ini/10.8.97
% test to catch a bug in parse 1.0

% operator preconditions
move(X,Y,T):- 
	time(T),
	moveable(X),
	block(Y),
	smaller(X,Y),
        on_something(X,T),
        on_something(Y,T),
	not blocked_move(X,Y,T).

% requires that move parameters must be clear

blocked_move(X,Y,T) :-
	time(T),
	moveable(X),
	block(Y),
	smaller(X,Y),
	moveable(Z),
       	smaller(Z,Y),
        on(Z,Y,T).

blocked_move(X,Y,T) :-
	time(T),
	moveable(X),
	block(Y),
	smaller(X,Y),
	moveable(Z),
       	smaller(Z,X),
        on(Z,X,T).


% enforces one move operation at a time 
blocked_move(X,Y,T) :- 
	time(T),
	moveable(X),
	block(Y),
	smaller(X,Y),
	move(X2,Y2,T),
      	not eq_block(Y,Y2).
blocked_move(X,Y,T) :-
	time(T),
	moveable(X),
	block(Y),
	smaller(X,Y),
        move(X2,Y2,T),
      	not eq_block(X,X2).
% stop applying operators when the goal has been reached
blocked_move(X,Y,T):- 
	time(T),
	moveable(X),
	block(Y),
        smaller(X,Y),
        goal(T).

% operator effects
on(X,Y,T2) :- 
	prevtime(T2,T1),
	moveable(X),
	block(Y),
	smaller(X,Y),
	move(X,Y,T1).



% frame axioms
on(X,Y,T2) :- 
	prevtime(T2,T1),
        moveable(X), 
        block(Y), 
	smaller(X,Y),
        on(X,Y,T1),
        not change_on(X,Y,T1).

change_on(X,Z,T) :- 
	time(T), 
        moveable(X), 
        block(Y), 
	smaller(X,Y),
        block(Z), 
	smaller(X,Z),
        move(X,Y,T), 
        on(X,Z,T).

% defined predicates
% on_something
on_something(X,T) :-
	time(T),
	moveable(X),
	block(Y),
        smaller(X,Y),
	on(X,Y,T).
% X is peg
on_something(X,T) :-
	time(T),
	block(X), 
	not moveable(X).

% pruning rules
% do not go back to the previous configuration
inco :- 
	prevtime(T2,T1),
	move(X,Y,T1), 
        on(X,Z,T1),
	move(X,Z,T2).

% do not move the same object twice in a row
inco :- 
	prevtime(T2,T1),
	move(X,Y,T1), 
        not eq_block(Y,Z),
	move(X,Z,T2).

% initial situation
on(b,a,t0).
on(a,peg1,t0).

% goal 

goal(T) :- time(T), on(a,peg3,T), on(b,a,T).

goal(T2) :- prevtime(T2,T1), goal(T1).

goal :- time(T), goal(T).

compute 1 {goal, not inco}


% facts
eq_block(X,X) :- block(X).

block(a).
block(b).

block(peg1).
block(peg2).
block(peg3).

moveable(a).
moveable(b).

time(t0).
time(t1).
time(t2).
time(t3).

prevtime(t1,t0).
prevtime(t2,t1).
prevtime(t3,t2).

smaller(b,a).
smaller(b,peg1).
smaller(b,peg2).
smaller(b,peg3).

smaller(a,peg1).
smaller(a,peg2).
smaller(a,peg3).


